@@ -47,6 +47,7 @@ private slots:
     void storageBufferQualifiers();
     void multiview();
     void mediump();
+    void geometryShader();
 };
 
 void tst_QShaderBaker::initTestCase()
@@ -318,6 +319,10 @@ void tst_QShaderBaker::compileError()
 void tst_QShaderBaker::translateError()
 {
     // assume the shader here fails in SPIRV-Cross with "cbuffer cannot be expressed with either HLSL packing layout or packoffset"
+    //
+    // This has been improved in SPIRV-Cross, presumably by 0ac65a353e94e9756e8b89fa74e953c482a915e0, so skip this test.
+    QSKIP("Not applicable with updated SPIRV-Cross");
+
     QShaderBaker baker;
     baker.setSourceFileName(QLatin1String(":/data/hlsl_cbuf_error.frag"));
     baker.setGeneratedShaderVariants({ QShader::StandardShader });
@@ -1450,6 +1455,24 @@ void tst_QShaderBaker::mediump()
     QVERIFY(shader.shader().contains("precision highp float;"));
 }
 
+void tst_QShaderBaker::geometryShader()
+{
+    QShaderBaker baker;
+    baker.setSourceFileName(QLatin1String(":/data/voxelize.geom"));
+    baker.setGeneratedShaderVariants({ QShader::StandardShader });
+    QList<QShaderBaker::GeneratedShader> targets;
+    targets.append({ QShader::SpirvShader, QShaderVersion(100) });
+    targets.append({ QShader::GlslShader, QShaderVersion(410) });
+    targets.append({ QShader::GlslShader, QShaderVersion(320, QShaderVersion::GlslEs) });
+    targets.append({ QShader::HlslShader, QShaderVersion(50) });
+    baker.setGeneratedShaders(targets);
+    QShader s = baker.bake();
+    QVERIFY(s.isValid());
+    QVERIFY(s.stage() == QShader::GeometryStage);
+    QVERIFY(baker.errorMessage().isEmpty());
+    QCOMPARE(s.availableShaders().size(), 4);
+}
+
 
 #include <tst_qshaderbaker.moc>
 QTEST_MAIN(tst_QShaderBaker)
